from redgifs.api import API

def test_routes():
    api = API().login()
    # assert api.get_feeds() # removed on v2.0.0
    assert api.get_tags()
    assert api.get_top_this_week()
    assert api.get_trending_gifs()
    assert api.get_trending_images()
    assert api.get_trending_tags()
    assert api.fetch_tag_suggestions('mia malkova')



from redgifs import API, Order

api = API().login()

# GIFS
def test_order_recent():
    r = api.search('big ass', order=Order.recent)
    assert r

# def test_order_followers():
#     r = api.search('big dick', order=Order.followers)
#     assert r

def test_order_best():
    r = api.search('big tits', order=Order.best)
    assert r

def test_order_trending():
    r = api.search('hitomi tanaka', order=Order.trending)
    assert r

def test_order_top28():
    r = api.search('hitomi tanaka', order=Order.top28)
    assert r

def test_order_new():
    r = api.search('ava addams', order=Order.new)
    assert r

def test_order_latest():
    r = api.search('violet myers', order=Order.latest)
    assert r

def test_order_oldest():
    r = api.search('mia khalifa', order=Order.oldest)
    assert r

# ----------------------------- #

# IMAGES
def test_order_recent_i():
    r = api.search_image('big ass', order=Order.recent)
    assert r

# def test_order_followers_i():
#     r = api.search_image('big dick', order=Order.followers)
#     assert r

def test_order_best_i():
    r = api.search_image('big tits', order=Order.best)
    assert r

def test_order_trending_i():
    r = api.search_image('hitomi tanaka', order=Order.trending)
    assert r

def test_order_top28_i():
    r = api.search_image('hitomi tanaka', order=Order.top28)
    assert r

def test_order_new_i():
    r = api.search_image('ava addams', order=Order.new)
    assert r

def test_order_latest_i():
    r = api.search_image('violet myers', order=Order.latest)
    assert r

def test_order_oldest_i():
    r = api.search_image('mia khalifa', order=Order.oldest)
    assert r



from typing import List, TypedDict


class TagInfo(TypedDict):
    name: str
    count: int


class TagsResponse(TypedDict):
    tags: List[TagInfo]


class TagSuggestion(TypedDict):
    gifs: int
    text: str
    type: str



from typing import List, Optional, TypedDict

from .niches import NichesInfo
from .user import UserInfo

class MediaInfo(TypedDict):
    sd: str
    hd: str
    gif: str
    poster: str
    thumbnail: str
    vthumbnail: Optional[str]


class GifInfo(TypedDict):
    id: str
    client_id: Optional[str]
    createDate: int
    hasAudio: bool
    width: int
    height: int
    likes: int
    tags: List[str]
    verified: bool
    views: Optional[int]
    duration: float
    published: bool
    type: int # Literal[1,2]
    urls: MediaInfo
    userName: str
    avgColor: str
    gallery: str


class GetGifResponse(TypedDict):
    gif: GifInfo
    user: Optional[UserInfo]


class BaseGifResponse(TypedDict):
    page: int
    pages: int
    total: int
    gifs: List[GifInfo]
    users: List[UserInfo]
    niches: List[NichesInfo]
    tags: List[str]


class GifResponse(BaseGifResponse):
    pass



"""
The MIT License (MIT)

Copyright (c) 2022-present scrazzz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import io
import os
import sys
import logging
from urllib.parse import quote
from typing import TYPE_CHECKING, Any, ClassVar, Coroutine, Dict, List, NamedTuple, Optional, TypeVar, Union

import requests
import aiohttp
import yarl

from . import __version__
from .errors import HTTPException
from .enums import Order, MediaType
from .utils import strip_ip

__all__ = ('ProxyAuth',)

_log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from redgifs.types.gif import GetGifResponse, GifResponse
    from redgifs.types.image import ImageResponse, TrendingImagesResponse
    from redgifs.types.tags import TagsResponse, TagSuggestion
    from redgifs.types.user import CreatorResponse, CreatorsResponse

    T = TypeVar('T')
    Response = Coroutine[Any, Any, T]

class Route:
    BASE: ClassVar[str] = "https://api.redgifs.com"

    def __init__(self, method: str, path: str, **parameters: Any) -> None:
        self.method: str = method
        self.path: str = path
        url: str = self.BASE + self.path
        if parameters:
            url = url.format_map({k: quote(v) if isinstance(v, str) else v for k, v in parameters.items()})
        self.url: str = url

class ProxyAuth(NamedTuple):
    """
    username: :class:`str`
        The username.
    password: :class:`str`
        The password.
    """
    username: str
    password: str

class HTTP:
    def __init__(
        self,
        session: Optional[requests.Session] = None,
        *,
        proxy: Optional[str] = None,
        proxy_auth: Optional[ProxyAuth] = None
    ) -> None:

        if session is not None and session is not isinstance(session, requests.Session):
            raise RuntimeError("session is not of type requests.Session")

        self.__session: requests.Session = session or requests.Session()
        self.headers: Dict[str, str] = {
            'User-Agent': f'redgifs (https://github.com/scrazzz/redgifs {__version__}) Python/{sys.version[:3]}'
        }
        self.proxy: Optional[yarl.URL] = yarl.URL(proxy) if proxy else None
        self.proxy_auth: Optional[ProxyAuth] = proxy_auth

        # Set proxy
        if self.proxy:
            self._proxy = {self.proxy.scheme: str(self.proxy)}
        else:
            self._proxy = None

        # Set proxy auth
        if self.proxy_auth and self._proxy:
            self._proxy_auth = (self.proxy_auth.username, self.proxy_auth.password)
        else:
            self._proxy_auth = None

    def request(self, route: Route, **kwargs: Any) -> Any:
        url: str = route.url
        method: str = route.method
        r: requests.Response = self.__session.request(
            method, url, headers=self.headers, proxies=self._proxy, auth=self._proxy_auth, timeout=60.0, **kwargs
        )
        _log.debug(f'{method} {url} returned code: {r.status_code}')
        js = r.json()
        if r.status_code == 200:
            _log.debug(f'{method} {url} received: {js}')
            return js
        else:
            raise HTTPException(r, js)

    def close(self) -> None:
        self.__session.close()

    # TODO: Implement OAuth login support
    def login(self, username: Optional[str] = None, password: Optional[str] = None) -> None:
        if (username and password) is None:
            temp_token = self.get_temp_token()['token']
            self.headers['authorization'] = f'Bearer {temp_token}'
        else:
            raise NotImplementedError

    def get_temp_token(self):
        return self.request(Route('GET', '/v2/auth/temporary'))

    # GIF methods

    def get_tags(self, **params: Any) -> TagsResponse:
        return self.request(Route('GET', '/v1/tags'), **params)

    def get_gif(self, id: str, **params: Any) -> GetGifResponse:
        return self.request(Route('GET', '/v2/gifs/{id}', id=id), **params)

    def search(self, search_text: str, order: Order, count: int, page: int, **params: Any) -> GifResponse:
        r = Route(
            'GET', '/v2/gifs/search?search_text={search_text}&order={order}&count={count}&page={page}',
            search_text=search_text, order=order.value, count=count, page=page
        )
        return self.request(r, **params)

    def get_top_this_week(self, count: int, page: int, type: MediaType, **params) -> GifResponse:
        r = Route(
            'GET', '/v2/gifs/search?order=top7&count={count}&page={page}&type={type}',
            count=count, page=page, type=type.value
        )
        return self.request(r, **params)

    # User/Creator methods

    def search_creators(
        self,
        page: int,
        order: Order,
        verified: bool,
        tags: Optional[List[str]],
        **params: Any
    ) -> CreatorsResponse:
        url = '/v1/creators/search?page={page}&order={order}'
        if verified:
            url += '&verified={verified}'
        if tags:
            url += '&tags={tags}'
            r = Route(
                'GET', url,
                page=page, order=order.value, verified='y' if verified else 'n', tags=','.join(t for t in tags)
            )
            return self.request(r, **params)
        else:
            r = Route(
                'GET', url,
                page=page, order=order.value, verified='y' if verified else 'n'
            )
            return self.request(r, **params)

    def search_creator(self, username: str, page: int, count: int, order: Order, type: MediaType, **params) -> CreatorResponse:
        r = Route(
            'GET', '/v2/users/{username}/search?page={page}&count={count}&order={order}&type={type}',
            username=username, page=page, count=count, order=order.value, type=type.value
        )
        return self.request(r, **params)

    def get_trending_gifs(self) -> GifResponse:
        r = Route('GET', '/v2/explore/trending-gifs')
        return self.request(r)

    # Pic methods

    def search_image(self, search_text: str, order: Order, count: int, page: int, **params: Any) -> ImageResponse:
        r = Route(
            'GET', '/v2/gifs/search?search_text={search_text}&order={order}&count={count}&page={page}&type=i',
            search_text=search_text, order=order.value, count=count, page=page
        )
        return self.request(r, **params)

    def get_trending_images(self) -> TrendingImagesResponse:
        r = Route('GET', '/v2/explore/trending-images')
        return self.request(r)

    # Tag methods

    def get_trending_tags(self) -> TagsResponse:
        r = Route('GET', '/v2/search/trending')
        return self.request(r)

    def get_tag_suggestions(self, query: str) -> List[TagSuggestion]:
        r = Route('GET', '/v2/search/suggest?query={query}', query=query)
        return self.request(r)

    # download

    def download(self, url: str, fp: Union[str, bytes, os.PathLike[Any], io.BufferedIOBase]) -> Union[Coroutine[Any, Any, int], int]:
        """A friendly method to download a RedGifs media."""

        yarl_url = yarl.URL(url)
        str_url = str(yarl_url)

        def dl(url: str) -> int:
            r = self.__session.get(url, headers = self.headers)
            _log.debug(f'GET {url} returned code: {r.status_code}')

            if r.status_code == 404:
                raise HTTPException(r, r.json())

            content_type = r.headers['Content-Type']
            if content_type not in ['video/mp4', 'image/jpeg']:
                _log.error(f'GET {url} returned improper content-type: {content_type}')
                raise TypeError(f'"{url}" returned invalid content type for downloading: {content_type}')

            data = r.content
            if isinstance(fp, io.BufferedIOBase):
                return fp.write(data)
            else:
                with open(fp, 'wb') as f:
                    return f.write(data)

        if (yarl_url.host is not None and 'redgifs.com' in yarl_url.host):
            if 'watch' in yarl_url.path:
                id = yarl_url.path.strip('/watch/')
                hd_url = self.get_gif(id)['gif']['urls']['hd']
                return dl(hd_url)
            else:
                return dl(str_url)

        raise TypeError(f'"{strip_ip(str_url)}" is not a valid RedGifs URL')


class AsyncHttp:
    def __init__(
        self,
        session: Optional[aiohttp.ClientSession] = None,
        *,
        proxy: Optional[str] = None,
        proxy_auth: Optional[ProxyAuth] = None
    ) -> None:

        if session is not None and session is not isinstance(session, aiohttp.ClientSession):
            raise RuntimeError("session is not of type aiohttp.ClientSession")

        self.__session: aiohttp.ClientSession = session or aiohttp.ClientSession()
        self.headers: Dict[str, str] = {
            'User-Agent': f'redgifs (https://github.com/scrazzz/redgifs {__version__}) Python/{sys.version[:3]}'
        }
        self.proxy: Optional[yarl.URL] = yarl.URL(proxy) if proxy else None
        self.proxy_auth: Optional[ProxyAuth] = proxy_auth

        # Set proxy auth
        if self.proxy_auth and self.proxy:
            self._proxy_auth = aiohttp.BasicAuth(self.proxy_auth.username, self.proxy_auth.password)
        else:
            self._proxy_auth = None

    async def request(self, route: Route, **kwargs: Any) -> Any:
        url: str = route.url
        method: str = route.method
        async with self.__session.request(
            method, url, headers=self.headers, proxy=str(self.proxy) if self.proxy else None, proxy_auth=self._proxy_auth, **kwargs
        ) as resp:
            _log.debug(f'{method} {url} returned code: {resp.status}')
            js = await resp.json()
            if resp.status == 200:
                _log.debug(f'{method} {url} received: {js}')
                return js
            else:
                raise HTTPException(resp, js)

    async def close(self) -> None:
        await self.__session.close()

    async def get_temp_token(self):
        return (await self.request(Route('GET', '/v2/auth/temporary')))

    # TODO: Implement OAuth login support
    async def login(self, username: Optional[str] = None, password: Optional[str] = None) -> None:
        if (username and password) is None:
            temp_token = await self.get_temp_token()
            self.headers['authorization'] = f'Bearer {temp_token["token"]}'
        else:
            raise NotImplementedError
    
    # GIF methods

    def get_tags(self, **params: Any) -> Response[TagsResponse]:
        return self.request(Route('GET', '/v1/tags'), **params)

    def get_gif(self, id: str, **params: Any) -> Response[GetGifResponse]:
        return self.request(Route('GET', '/v2/gifs/{id}', id=id), **params)

    def search(self, search_text: str, order: Order, count: int, page: int, **params: Any) -> Response[GifResponse]:
        r = Route(
            'GET', '/v2/gifs/search?search_text={search_text}&order={order}&count={count}&page={page}',
            search_text=search_text, order=order.value, count=count, page=page
        )
        return self.request(r, **params)

    def get_top_this_week(self, count: int, page: int, type: MediaType, **params) -> Response[GifResponse]:
        r = Route(
            'GET', '/v2/gifs/search?order=top7&count={count}&page={page}&type={type}',
            count=count, page=page, type=type.value
        )
        return self.request(r, **params)

    # User/Creator methods

    def search_creators(
        self,
        page: int,
        order: Order,
        verified: bool,
        tags: Optional[List[str]],
        **params: Any
    ) -> Response[CreatorsResponse]:
        url = '/v1/creators/search?page={page}&order={order}'
        if verified:
            url += '&verified={verified}'
        if tags:
            url += '&tags={tags}'
            r = Route(
                'GET', url,
                page=page, order=order.value, verified='y' if verified else 'n', tags=','.join(t for t in tags)
            )
            return self.request(r, **params)
        else:
            r = Route(
                'GET', url,
                page=page, order=order.value, verified='y' if verified else 'n'
            )
            return self.request(r, **params)

    def search_creator(self, username: str, page: int, count: int, order: Order, type: MediaType, **params) -> Response[CreatorResponse]:
        r = Route(
            'GET', '/v2/users/{username}/search?page={page}&count={count}&order={order}&type={type}',
            username=username, page=page, count=count, order=order.value, type=type.value
        )
        return self.request(r, **params)

    def get_trending_gifs(self) -> Response[GifResponse]:
        r = Route('GET', '/v2/explore/trending-gifs')
        return self.request(r)

    # Pic methods

    def search_image(self, search_text: str, order: Order, count: int, page: int, **params: Any) -> Response[ImageResponse]:
        r = Route(
            'GET', '/v2/gifs/search?search_text={search_text}&order={order}&count={count}&page={page}&type=i',
            search_text=search_text, order=order.value, count=count, page=page
        )
        return self.request(r, **params)

    def get_trending_images(self) -> Response[TrendingImagesResponse]:
        r = Route('GET', '/v2/explore/trending-images')
        return self.request(r)

    # Tag methods

    def get_trending_tags(self) -> Response[TagsResponse]:
        r = Route('GET', '/v2/search/trending')
        return self.request(r)

    def get_tag_suggestions(self, query: str) -> Response[List[TagSuggestion]]:
        r = Route('GET', '/v2/search/suggest?query={query}', query=query)
        return self.request(r)

    # download

    async def download(self, url: str, fp: Union[str, bytes, os.PathLike[Any], io.BufferedIOBase]) -> int:
        yarl_url = yarl.URL(url)
        str_url = str(yarl_url)

        async def dl(url: str) -> int:
            async with self.__session.get(url, headers = self.headers) as r:
                _log.debug(f'GET {str_url} returned code: {r.status}')

                content_type = r.headers['Content-Type']
                if content_type not in ['video/mp4', 'image/jpeg']:
                    _log.error(f'GET {url} returned improper content-type: {content_type}')
                    raise TypeError(f'"{url}" returned invalid content type for downloading: {content_type}')

                data = await r.read()
                if isinstance(fp, io.BufferedIOBase):
                    return (fp.write(data))
                else:
                    with open(fp, 'wb') as f:
                        return f.write(data)

        if (yarl_url.host is not None and 'redgifs.com' in yarl_url.host):
            if 'watch' in yarl_url.path:
                id = yarl_url.path.strip('/watch/')
                hd_url = (await self.get_gif(id))['gif']['urls']['hd']
                return (await dl(hd_url))
            else:
                return (await dl(str_url))

        raise TypeError(f'"{strip_ip(str_url)}" is not a valid RedGifs URL')



"""
The MIT License (MIT)

Copyright (c) 2022-present scrazzz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import requests
import aiohttp

from typing import Any, Dict, Optional, Union

class RedGifsError(BaseException):
    """Base class for all redgifs errors."""
    pass

class InvalidTag(RedGifsError):
    """Exception raised when no match was found for a tag.

    Attributes
    ----------
    tag: :class:`str`
        The tag that was searched for.
    """
    def __init__(self, tag: str):
        self.tag: str = tag
        super().__init__(f'Tag for "{tag}" was not found.')

class HTTPException(RedGifsError):
    """Exception raised when an HTTP Exception occurs.

    Attributes
    ----------
    response: Union[:class:`requests.Response`, :class:`aiohttp.ClientResponse`]
        The response of the failed HTTP request. It may be either :class:`requests.Response` or :class:`aiohttp.ClientResponse`.
    status: :class:`int`
        The status code of the HTTP request.
    error: :class:`str`
        The original error message from RedGifs.
    """

    def __init__(self, response: Union[requests.Response, aiohttp.ClientResponse], json: Optional[Union[Dict[str, Any], str]]):
        self.response: Any = response

        if isinstance(response, requests.Response):
            self.status = response.status_code
        elif isinstance(response, aiohttp.ClientResponse):
            self.status = response.status

        self.error: Optional[Union[Dict[str, Any], str]] = json
        if isinstance(json, dict):
            self.error = json.get('errorMessage') or json.get('error') or json.get('message')

        super().__init__(f'{self.status} (Error: {self.error})')



"""
The MIT License (MIT)

Copyright (c) 2022-present scrazzz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import re
import json
import pkgutil
import asyncio
import yarl
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Dict, List

from .models import GIF, URL, Image, User
from .const import REDGIFS_THUMBS_RE

if TYPE_CHECKING:
    from redgifs.types.gif import GifInfo
    from redgifs.types.image import ImageInfo
    from redgifs.types.user import UserInfo

def to_web_url(id_or_url: str, use_regex: bool = False) -> str:
    if not use_regex:
        return f'https://redgifs.com/watch/{id_or_url.lower()}'

    match = re.match(REDGIFS_THUMBS_RE, id_or_url)
    if not match:
        return ''

    try:
        id = match.group('id')
        return f'https://redgifs.com/watch/{id.lower()}'
    except IndexError:
        return ''

def strip_ip(url: str) -> str:
    u = yarl.URL(url)
    if u.query.get('for'):
        return str(u % {'for': 'REDACTED'})
    return url

def build_file_url(url: str) -> str:
    # use the 'sd' url here
    u = yarl.URL(url)
    filename = u.path.replace('/', '').replace('-mobile.mp4', '')
    return f'https://api.redgifs.com/v2/gifs/{filename.lower()}/files/{filename}.mp4'

def to_embed_url(sd_url: str) -> str:
    u = yarl.URL(sd_url)
    filename = u.path.replace('/', '').replace('-mobile.mp4', '.mp4')
    return f'https://api.redgifs.com/v2/embed/discord?name={filename}'

def _read_tags_json() -> Dict[str, str]:
    file_ = pkgutil.get_data(__name__, 'tags.json')
    return json.loads(file_) # type: ignore - file_ will never be None

async def _async_read_tags_json() -> Dict[str, str]:
    r = await asyncio.get_event_loop().run_in_executor(None, _read_tags_json)
    return r

def _gifs_iter(gifs: List[GifInfo]) -> List[GIF]:
    return [
        GIF(
            id=g['id'],
            create_date=datetime.fromtimestamp(g['createDate'], tz=timezone.utc),
            has_audio=g['hasAudio'],
            width=g['width'],
            height=g['height'],
            likes=g['likes'],
            tags=g['tags'],
            verified=g['verified'],
            views=g['views'],
            duration=int(g['duration']) if g['duration'] is not None else g['duration'],
            published=g['published'],
            urls=URL(
                sd=g['urls']['sd'],
                hd=g['urls']['hd'],
                poster=g['urls'].get('poster'),
                thumbnail=g['urls']['thumbnail'],
                vthumbnail=g['urls']['vthumbnail'],
                web_url=to_web_url(g['id']),
                file_url=build_file_url(g['urls']['sd']),
                embed_url=to_embed_url(g['urls']['sd']),
            ),
            username=g['userName'],
            type=g['type'],
            avg_color=g['avgColor'],
        )
        for g in gifs
    ]

def _images_iter(images: List[ImageInfo]) -> List[Image]:
    return [
        Image(
            id=i['id'],
            create_date=datetime.fromtimestamp(i['createDate'], tz=timezone.utc),
            width=i['width'],
            height=i['height'],
            likes=i['likes'],
            tags=i['tags'],
            verified=i['verified'],
            views=i['views'],
            published=i['published'],
            urls=URL(
                sd=i['urls']['sd'],
                hd=i['urls']['hd'],
                poster=i['urls'].get('poster'),
                thumbnail=i['urls']['thumbnail'],
                vthumbnail=i['urls']['vthumbnail'],
                web_url=to_web_url(i['id']),
                file_url=None,
                embed_url=None,
            ),
            username=i['userName'],
            type=i['type'],
            avg_color=i['avgColor'],
        )
        for i in images
    ]

def _users_iter(users: List[UserInfo]) -> List[User]:
    return [
        User(
            # I only had this occurrence once where redgifs did not
            # send the response properly and messed up the entire JSON
            # response, this is why I have used dict.get() here.
            creation_time=datetime.fromtimestamp(user.get('creationtime'), tz=timezone.utc) if user.get('creationtime') is not None else None,
            description=user.get('description'),
            followers=user.get('followers'),
            following=user.get('following'),
            gifs=user.get('gifs'),
            name=user.get('name'),
            profile_image_url=user.get('profileImageUrl'),
            profile_url=user.get('profileUrl'),
            published_collections=user.get('publishedCollections'),
            status=user.get('status'),
            published_gifs=user.get('publishedGifs'),
            subscription=user.get('subscription'),
            url=user.get('url'),
            username=user.get('username'),
            verified=user.get('verified'),
            views=user.get('views'),
            poster=user.get('poster'),
            preview=user.get('preview'),
            thumbnail=user.get('thumbnail'),
            links=user.get('links')
        )
        for user in users
    ]



"""
The MIT License (MIT)

Copyright (c) 2022-present scrazzz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import difflib
from random import choice, choices
from typing import Dict, List, Literal, overload

from .utils import _read_tags_json
from .errors import InvalidTag

__all__ = ('Tags',)

class Tags:
    tags_mapping: Dict[str, str] = {}

    def _set(self, tags: Dict[str, str]) -> None:
        self.tags_mapping = tags

    def _clear(self) -> None:
        self.tags_mapping = {}

    def search(self, tag: str) -> List[str]:
        """
        Search for a specific RedGifs tag.

        .. note::

            This method is synchronous, a.k.a "blocking", when used alone.

        Parameters
        ----------
        tag: :class:`str`
            The tag name to search.

        Returns
        -------
        List[:class:`str`] - A list of tag names that are similar.
        """

        # If this method is being used alone, then `tags_mapping` needs to be set
        if len(self.tags_mapping) == 0:
            self._set(_read_tags_json())

        # Try to do a quick lookup, if it fails then go for difflib method
        try:
            return [self.tags_mapping[tag]]
        except KeyError:
            results = difflib.get_close_matches(tag.title(), self.tags_mapping.values())

            if len(results) == 0:
                raise InvalidTag(tag)

            return results

    @overload
    def random(self, count: Literal[1] = ...) -> str:
        ...

    @overload
    def random(self, count: int) -> List[str]:
        ...

    def random(self, count: int = 1):
        """
        Search for random RedGifs tags.

        Parameters
        ----------
        count: :class:`int`
            The amount of tags to return.

        Returns
        -------
        Union[:class:`str`, List[:class:`str`]]
            If the ``count`` specified is ``1`` then a single random tag is returned
            or else a list of tags are returned.
        """
        return choices(list(self.tags_mapping.values()), k=count) if count != 1 \
        else choice(list(self.tags_mapping.values()))



"""
The MIT License (MIT)

Copyright (c) 2022-present scrazzz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import logging
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from .enums import MediaType
from .utils import _users_iter, build_file_url, to_embed_url, to_web_url
from .models import GIF, URL, CreatorResult, Image, User, SearchResult, CreatorsResult

if TYPE_CHECKING:
    from redgifs.types.gif import GifResponse
    from redgifs.types.image import ImageResponse
    from redgifs.types.user import CreatorResponse, CreatorsResponse

_log = logging.getLogger(__name__)

# For GIFs
def parse_search(searched_for: str, json: GifResponse, media_type: MediaType) -> SearchResult:
    _log.debug('Using `parse_search` for: {searched_for}')
    json_gifs = json['gifs']
    users = json['users']
    return SearchResult(
        searched_for=searched_for,
        page=json['page'],
        pages=json['pages'],
        total=json['total'],
        images=[
            Image(
                id=img['id'],
                create_date=datetime.fromtimestamp(img['createDate'], tz=timezone.utc),
                width=img['width'],
                height=img['height'],
                likes=img['likes'],
                tags=img['tags'],
                verified=img['verified'],
                views=img['views'],
                published=img['published'],
                urls=URL(
                    sd=img['urls']['sd'],
                    hd=img['urls']['hd'],
                    poster=img['urls']['poster'],
                    thumbnail=img['urls']['thumbnail'],
                    vthumbnail=img['urls']['vthumbnail'],
                    web_url=to_web_url(img['id']),
                    file_url=None,
                    embed_url=to_embed_url(img['urls']['hd']),
                ),
                username=img['userName'],
                type=img['type'],
                avg_color=img['avgColor'],
            ) for img in json['gifs'] if media_type == MediaType.IMAGE
        ],
        gifs=[
            GIF(
                id=gif['id'],
                create_date=datetime.fromtimestamp(gif['createDate'], tz=timezone.utc),
                has_audio=gif['hasAudio'],
                width=gif['width'],
                height=gif['height'],
                likes=gif['likes'],
                tags=gif['tags'],
                verified=gif['verified'],
                views=gif['views'],
                duration=int(gif['duration']) if gif['duration'] is not None else gif['duration'],
                published=gif['published'],
                urls=URL(
                    sd=gif['urls']['sd'],
                    hd=gif['urls']['hd'],
                    poster=gif['urls']['poster'],
                    thumbnail=gif['urls']['thumbnail'],
                    vthumbnail=gif['urls']['vthumbnail'],
                    web_url=to_web_url(gif['id']),
                    file_url=build_file_url(gif['urls']['sd']),
                    embed_url=to_embed_url(gif['urls']['sd']),
                ),
                username=gif['userName'],
                type=gif['type'],
                avg_color=gif['avgColor'],
            )
            for gif in json_gifs if media_type == MediaType.GIF
        ],
        users=[
            User(
                # I only had this occurrence once where redgifs did not
                # send the response properly and messed up the entire JSON
                # response, this is why I have used dict.get() here.
                creation_time=datetime.fromtimestamp(user.get('creationtime'), tz=timezone.utc)
                if user.get('creationtime') is not None else None,
                description=user.get('description'),
                followers=user.get('followers'),
                following=user.get('following'),
                gifs=user.get('gifs'),
                name=user.get('name'),
                profile_image_url=user.get('profileImageUrl'),
                profile_url=user.get('profileUrl'),
                published_collections=user.get('publishedCollections'),
                status=user.get('status'),
                published_gifs=user.get('publishedGifs'),
                subscription=user.get('subscription'),
                url=user.get('url'),
                username=user.get('username'),
                verified=user.get('verified'),
                views=user.get('views'),
                poster=user.get('poster'),
                preview=user.get('preview'),
                thumbnail=user.get('thumbnail'),
                links=user.get('links'),
            )
            for user in users
        ],
        tags=json['tags'],
    )

# For images
def parse_search_image(searched_for: str, json: ImageResponse) -> SearchResult:
    _log.debug('Using `parse_search` for: {searched_for}')
    json_gifs = json['gifs']
    users = json['users']
    return SearchResult(
        searched_for=searched_for,
        page=json['page'],
        pages=json['pages'],
        total=json['total'],
        gifs=None,
        images=[
            Image(
                id=gif['id'],
                create_date=datetime.fromtimestamp(gif['createDate'], tz=timezone.utc),
                width=gif['width'],
                height=gif['height'],
                likes=gif['likes'],
                tags=gif['tags'],
                verified=gif['verified'],
                views=gif['views'],
                published=gif['published'],
                urls=URL(
                    sd=gif['urls']['sd'],
                    hd=gif['urls']['hd'],
                    poster=gif['urls']['poster'],
                    thumbnail=gif['urls']['thumbnail'],
                    vthumbnail=gif['urls']['vthumbnail'],
                    web_url=to_web_url(gif['id']),
                    file_url=None,
                    embed_url=gif['urls']['hd'],
                ),
                username=gif['userName'],
                type=gif['type'],
                avg_color=gif['avgColor'],
            )
            for gif in json_gifs
        ],
        users=_users_iter(users),
        tags=json['tags'],
    )

def parse_creators(json: CreatorsResponse) -> CreatorsResult:
    _log.debug('Using `parse_creators`')
    items = json['items']
    return CreatorsResult(
        items=_users_iter(items),
        pages=json['pages'],
        page=json['page'],
        total=json['total'],
    )

def parse_creator(json: CreatorResponse, media_type: MediaType) -> CreatorResult:
    _log.debug('Using `parse_creator`')
    user = json['users'][0]
    return CreatorResult(
        creator=User(
            creation_time=datetime.fromtimestamp(user['creationtime'], tz=timezone.utc),
            description=user['description'],
            followers=user['followers'],
            following=user['following'],
            gifs=user['gifs'],
            name=user['name'],
            profile_image_url=user['profileImageUrl'],
            profile_url=user['profileUrl'],
            published_collections=user['publishedCollections'],
            status=user['status'],
            published_gifs=user['publishedGifs'],
            subscription=user['subscription'],
            url=user['url'],
            username=user['username'],
            verified=user['verified'],
            views=user['views'],
            poster=user.get('poster'),
            preview=user.get('preview'),
            thumbnail=user.get('thumbnail'),
            links=user.get('links'),
        ),
        page=json['page'],
        pages=json['pages'],
        total=json['total'],
        gifs=[
            GIF(
                id=gif['id'],
                create_date=datetime.fromtimestamp(gif['createDate'], tz=timezone.utc),
                has_audio=gif['hasAudio'], # type: ignore - We aren't setting values for ImageInfo
                width=gif['width'],
                height=gif['height'],
                likes=gif['likes'],
                tags=gif['tags'],
                verified=gif['verified'],
                views=gif['views'],
                duration=gif['duration'], # type: ignore - We aren't setting values for ImageInfo
                published=gif['published'],
                urls=URL(
                    sd=gif['urls']['sd'],
                    hd=gif['urls']['hd'],
                    poster=gif['urls']['poster'],
                    thumbnail=gif['urls']['thumbnail'],
                    vthumbnail=gif['urls']['vthumbnail'],
                    web_url=to_web_url(gif['id']),
                    file_url=build_file_url(gif['urls']['sd']),
                    embed_url=to_embed_url(gif['urls']['sd']),
                ),
                username=gif['userName'],
                type=gif['type'],
                avg_color=gif['avgColor'],
            )
            for gif in json['gifs'] if media_type == MediaType.GIF
        ],
        images=[
            Image(
                id=img['id'],
                create_date=datetime.fromtimestamp(img['createDate'], tz=timezone.utc),
                width=img['width'],
                height=img['height'],
                likes=img['likes'],
                tags=img['tags'],
                verified=img['verified'],
                views=img['views'],
                published=img['published'],
                urls=URL(
                    sd=img['urls']['sd'],
                    hd=img['urls']['hd'],
                    poster=img['urls']['poster'],
                    thumbnail=img['urls']['thumbnail'],
                    vthumbnail=img['urls']['vthumbnail'],
                    web_url=to_web_url(img['id']),
                    file_url=None,
                    embed_url=img['urls']['hd'],
                ),
                username=img['userName'],
                type=img['type'],
                avg_color=img['avgColor'],
            ) for img in json['gifs'] if media_type == MediaType.IMAGE # RedGifs return the key for this data as "gifs" even though it's an image...
        ]
    )



from typing import List, Optional, TypedDict, Union

from .image import ImageInfo
from .gif import GifInfo
from .niches import NichesInfo


class UserInfo(TypedDict):
    creationtime: int
    description: Optional[str]
    followers: int
    following: int
    gifs: int
    name: str
    profileImageUrl: Optional[str]
    profileUrl: str
    publishedCollections: Optional[int]
    publishedGifs: int
    status: Optional[str]
    subscription: int
    url: str
    username: str
    verified: bool
    views: int
    poster: Optional[str]
    preview: Optional[str]
    thumbnail: Optional[str]
    likes: Optional[int]


class CreatorInfo(UserInfo):
    pass


# NOTE: See type of `gifs` field
class CreatorResponse(TypedDict):
    page: int
    pages: int
    total: int
    gifs: Union[List[GifInfo], List[ImageInfo]]
    users: List[UserInfo]
    niches: List[NichesInfo]
    tags: List[str]


class CreatorsResponse(TypedDict):
    page: int
    pages: int
    total: int
    items: List[UserInfo]



import pytest

from redgifs import Tags

@pytest.mark.parametrize(
    "test_input, expected",
    [
        ('tits', 'Tits'),
        ('ass', 'Ass'),
        #('cum', 'Cum'),
        ('americam', 'American'),
        ('japanes', 'Japanese'),
        ('hitomi tanaka', 'Hitomi Tanaka'),
        ('big dick', 'Big Dick'),
        ('hige tits', 'Huge Tits'),
        ('alexa pearl', 'Alexa Pearl'),
        ('ava adams', 'Ava Addams'),
    ]
)
def test_tags_search(test_input, expected):
    searched = Tags().search(test_input)[0]
    assert searched == expected


from typing import List, Optional, TypedDict

from .gif import MediaInfo
from .niches import NichesInfo
from .user import UserInfo


class ImageInfo(TypedDict):
    id: str
    client_id: Optional[str]
    createDate: int
    # hasAudio: bool
    width: int
    height: int
    likes: int
    tags: List[str]
    verified: bool
    views: Optional[int]
    published: bool
    type: int # Literal[1,2]
    urls: MediaInfo
    userName: str
    avgColor: str
    # gallery: str
    niches: List[str]
    sexuality: Optional[List[str]]


# NOTE: even though this is an image, the field for this is `gifs`
# NOTE: see the type of `gifs` key.
class BaseImageResponse(TypedDict):
    page: int
    pages: int
    total: int
    gifs: List[ImageInfo]
    users: List[UserInfo]
    niches: List[NichesInfo]
    tags: List[str]


class TrendingImagesResponse(BaseImageResponse):
    pass


class ImageResponse(BaseImageResponse):
    pass



"""
The MIT License (MIT)

Copyright (c) 2022-present scrazzz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import io
import os
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, List, Optional, Union

import requests

from .tags import Tags
from .enums import Order, MediaType
from .http import HTTP, ProxyAuth
from .utils import _read_tags_json, build_file_url, _gifs_iter, _images_iter, to_embed_url, to_web_url
from .parser import parse_creator, parse_search, parse_creators, parse_search_image
from .models import URL, GIF, CreatorResult, Image, SearchResult, CreatorsResult, TagSuggestion

if TYPE_CHECKING:
    from redgifs.types.tags import TagInfo

__all__ = ('API',)

class API:
    """The API Instance to get information from the RedGifs API.

    .. note::

        If you are using this library in an asynchronous code,
        you should be referencing :ref:`Async API` section.

    Parameters
    ----------
    session: Optional[:class:`requests.Session`]
        A session object that can be provided to do the requests.
        If not provided, a new session object is created.
    proxy: Optional[:class:`str`]
        A valid proxy URL.
    proxy_auth: Optional[:class:`.ProxyAuth`]
        The proxy auth to provide if the proxy requires it.
    """

    def __init__(
        self,
        session: Optional[requests.Session] = None,
        *,
        proxy: Optional[str] = None,
        proxy_auth: Optional[ProxyAuth] = None
    ) -> None:
        self.http: HTTP = HTTP(session, proxy=proxy, proxy_auth=proxy_auth)
        self._tags = Tags()

    def login(self) -> 'API':
        """
        A method to login to RedGifs with a temporary token.
        You must use this method after initialising the :class:`API <redgifs.API>` class 
        for this library to function properly.

        Returns
        -------
        :class:`API` - The properly initialised API class.
        """
        self.http.login()
        return self

    def get_tags(self) -> List[TagInfo]:
        """Get all available RedGifs Tags.
        
        Returns
        -------
        ``List[Dict[str, Union[str, int]]]``
        """
        return (self.http.get_tags()['tags'])
    
    def get_gif(self, id: str) -> GIF:
        """
        Get details of a single GIF using its ID.

        If the URL is ``https://redgifs.com/watch/abcxyz`` then the ID is ``abcxyz``.

        Parameters
        ----------
        id: :class:`str`
            The ID of the GIF.

        Returns
        -------
        :class:`.GIF` - The GIF's info.
        """

        json = self.http.get_gif(id)['gif']
        urls = json['urls']
        return GIF(
            id=json['id'],
            create_date=datetime.fromtimestamp(json['createDate'], tz=timezone.utc),
            has_audio=json['hasAudio'],
            width=json['width'],
            height=json['height'],
            likes=json['likes'],
            tags=json['tags'],
            verified=json['verified'],
            views=json['views'],
            duration=json['duration'],
            published=json['published'],
            urls=URL(
                sd=urls['sd'],
                hd=urls.get('hd'),
                poster=urls.get('poster'),
                thumbnail=urls['thumbnail'],
                vthumbnail=urls.get('vthumbnail'),
                web_url=to_web_url(json['id']),
                file_url=build_file_url(urls['sd']),
                embed_url=to_embed_url(urls['sd']),
            ),
            username=json['userName'],
            type=json['type'],
            avg_color=json['avgColor'],
        )
    
    def get_trending_gifs(self) -> List[GIF]:
        """
        Get the top 10 trending GIFs on RedGifs.

        Returns
        -------
        List[:class:`.GIF`]
        """
        r = self.http.get_trending_gifs()['gifs']
        return _gifs_iter(r)

    def get_trending_images(self) -> List[Image]:
        """
        Get the top 10 trending images on RedGifs.

        Returns
        -------
        List[:class:`.Image`]
        """
        r = self.http.get_trending_images()['gifs']
        return _images_iter(r)

    def get_trending_tags(self) -> List[TagInfo]:
        """Get the trending searches on RedGifs.

        Returns
        -------
        ``List[Dict[str, Union[str, int]]]``
            A list of dicts containing the tag name and count::

                [
                    {
                        "name": "r/CaughtPublic",
                        "count": 2034
                    },
                    {
                        "name": "Vintage",
                        "count": 19051
                    },
                    ...
                ]
        """
        resp = self.http.get_trending_tags()['tags']
        return resp

    def get_top_this_week(self, count: int = 30, page: int = 1, type: MediaType = MediaType.GIF) -> SearchResult:
        """Get media from "Top This Week" section.

        Parameters
        ----------
        count: :class:`int`
            The number of items to return.
        page: :class:`int`
            The items to return from given page number.
        type: :class:`.MediaType`
            The type of media to return.

        Returns
        -------
        :class:`.SearchResult` - Top this week results.
        """
        resp = self.http.get_top_this_week(count, page, type)
        return parse_search('TopThisWeek', resp, type)

    def fetch_tag_suggestions(self, query: str) -> List[TagSuggestion]:
        """Get tag suggestions from RedGifs.

        .. note::
        
            It's advised to use :func:`Tags.search()` to prevent multiple API calls to redgifs.com.

        Parameters
        ----------
        query: :class:`str`
            The tag name to look for.

        Returns
        -------
        List[:class:`.TagSuggestion`] - A list of ``TagSuggestion`` with tag name and count.
        """
        result = self.http.get_tag_suggestions(query)
        return [TagSuggestion(name=d['text'], count=d['gifs']) for d in result]

    def search(
        self,
        search_text: str,
        *,
        order: Order = Order.trending,
        count: int = 80,
        page: int = 1
    ) -> SearchResult:
        """
        Search for GIFs.

        Parameters
        ----------
        search_text: :class:`str`
            The type of GIFs to search for. Can be a string or an instance of :class:`Tags`.
        order: Optional[:class:`Order`]
            The order of the GIFs to return.
        count: Optional[:class:`int`]
            The amount of GIFs to return.
        page: Optional[:class:`int`]
            The page number of the GIFs to return.

        Returns
        -------
        :class:`.SearchResult` - The search result.
        """
        if len(self._tags.tags_mapping) == 0:
            tags = _read_tags_json()
            self._tags._set(tags)

        st = self._tags.search(search_text)[0]
        resp = self.http.search(st, order, count, page)
        return parse_search(st, resp, MediaType.GIF)

    search_gif = search

    def search_creators(
        self,
        *,
        page: int = 1,
        order: Order = Order.recent,
        verified: bool = False,
        tags: Optional[List[str]] = None,
    ) -> CreatorsResult:
        """
        Search for some RedGifs Creators.

        Parameters
        ----------
        page: Optional[:class:`int`]
            The result in page number to return.
        order: Optional[:class:`.Order`]
            The order of the creators to return.
        verified: Optional[:class:`bool`]
            Wheather to only return verified creators.
        tags: Optional[List[:class:`str`]]
            A list of tags to look for.
            Narrows down the results to creators that have contents with the given tags.

        Returns
        -------
        :class:`.CreatorsResult` - The search result.
        """
        resp = self.http.search_creators(page=page, order=order, verified=verified, tags=tags)
        return parse_creators(resp)

    def search_creator(
        self,
        username: str,
        *,
        page: int = 1,
        count: int = 80,
        order: Order = Order.recent,
        type: MediaType = MediaType.GIF,
    ) -> CreatorResult:
        """
        Search for a single RedGifs creator/user by username.

        Parameters
        ----------
        username: :class:`str`
            The username of the creator/user.
        page: :class:`int`
            The current page number of the creator/user's profile.
        count: :class:`int`
            The total amount of GIFs to return.
        order: :class:`Order`
            The order to return creator/user's GIFs.
        type: :class:`.MediaType`
            Whether to return image or GIF results. By default returns GIFs.

        Returns
        -------
        :class:`.CreatorResult` - The creator/user searched for.
        """
        resp = self.http.search_creator(username, page=page, count=count, order=order, type=type)
        return parse_creator(resp, type)

    search_user = search_creator

    def search_image(
        self,
        search_text: str,
        *,
        order: Order = Order.new,
        count: int = 80,
        page: int = 1
    ) -> SearchResult:
        """
        Search for images.

        Parameters
        ----------
        search_text: :class:`str`
            The images to search for. Can be a string or an instance of :class:`Tags <redgifs.Tags>`.
        order: Optional[:class:`.Order`]
            The order of the images to return.
        count: Optional[:class:`int`]
            The amount of images to return.
        page: Optional[:class:`int`]
            The page number of the images to return.

        Returns
        -------
        :class:`.SearchResult` - The search result.
        """
        # We are not going to use Tags.search() here because it doesn't matter
        # whatever the search_text is, this API endpoints provides images nonetheless.
        resp = self.http.search_image(search_text, order, count, page)
        return parse_search_image(search_text, resp)

    def download(self, url: str, fp: Union[str, bytes, os.PathLike[Any], io.BufferedIOBase]):
        """
        A friendly method to download a RedGifs media.

        Example:

        .. code-block:: python

            api = redgifs.API()
            api.login()
            hd_url = api.search("query").gifs[0].urls.hd
            api.download(hd_url, "video.mp4")

        .. note::
            
            You should use this method to download any media from RedGifs
            because RedGifs does validation on User-Agents and other params. 
            If you try to download it by using any other means, it will return 403 error.

        Parameters
        ----------
        url: :class:`str`
            A valid RedGifs URL.
        fp: Union[:class:`io.BufferedIOBase`, :class:`os.PathLike`]
            The file-like object to save this asset to or the filename
            to use. If a filename is passed then a file is created with that
            filename and used instead.
        """
        return self.http.download(url, fp)

    def close(self) -> None:
        """Closes the API session."""
        return self.http.close()



import pytest
import redgifs

@pytest.mark.parametrize(
    'query',
    [
        'hitomi',
        'angela white',
        'boobs',
        'ass'
    ]
)
def test_images_exist_in_result(query):
    api = redgifs.API()
    api.login()
    result = api.search_image(query)
    assert result.images



from redgifs import API

def test_attrs():
    api = API()
    api.login()
    r = api.search('slim')

    if r.gifs is not None:
        for gif in r.gifs:
            assert gif.id is not None
            assert gif.create_date
            assert isinstance(gif.has_audio, bool)
            assert gif.width and gif.height
            assert isinstance(gif.likes, int)
            assert gif.tags
            assert isinstance(gif.published, bool)
            assert gif.urls



import pytest

from redgifs import API

api = API()
api.login()

@pytest.mark.parametrize(
    "search_text, searched_for",
    [
        ('african', 'African'),
        ('ahegao', 'Ahegao'),
        ('ana lord', 'Ana Lorde'),
        ('sex', 'Sex'),
    ]
)
def test_search_with_search_text(search_text, searched_for):
    result = api.search(search_text)
    assert result.searched_for == searched_for


@pytest.mark.parametrize(
    "count, expected_count",
    [ (10, 10), (20, 20), (35, 35) ]
)
def test_search_with_count(count, expected_count):
    result = api.search('hitomi tanaka', count=count)
    assert result.gifs is not None and len(result.gifs) >= expected_count


@pytest.mark.parametrize(
    "page, expected_page",
    [ (2, 2), (3, 3), (6, 6) ]
)
def test_search_with_page(page, expected_page):
    result = api.search('big dick', page=page)
    assert result.page == expected_page



"""
The MIT License (MIT)

Copyright (c) 2022-present scrazzz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import datetime
from dataclasses import dataclass
from typing import Dict, List, Optional, TypedDict

@dataclass
class URL:
    """The different types of URLs.

    .. warning::

        ``sd``, ``hd``, ``poster``, ``thumbnail``, and ``vthumbnail`` leaks your IP address in the URL.
        If you want to display the URLs to the end user consider using ``web_url`` or ``embed_url`` instead.

    Attributes
    ----------
    sd: :class:`str`
        The sd URL of the media.
    hd: :class:`str`
        The hd URL of the media.
    poster: Optional[:class:`str`]
        The poster URL of the media.
    thumbnail: :class:`str`
        The thumbnail URL of the media.
    vthumbnail: Optional[:class:`str`]
        The vthumbnail URL of the media.
    web_url: :class:`str`
        The website URL of the media.
    file_url: Optional[:class:`str`]
        The file URL of the media.
    embed_url: Optional[:class:`str`]
        The embed URL of the media. This can be used to load the media without any restrictions.
    """

    __slots__ = ('sd', 'hd', 'poster', 'thumbnail', 'vthumbnail', 'web_url', 'file_url', 'embed_url')

    sd: str
    hd: str
    poster: Optional[str]
    thumbnail: str
    vthumbnail: Optional[str]
    web_url: str
    file_url: Optional[str]
    embed_url: Optional[str]

@dataclass
class GIF:
    # TODO: Document "type" and "avg_color"
    """The GIF returned from RedGifs.

    Attributes
    ----------
    id: :class:`str`
        The GIF's ID.
    create_date: Optional[:class:`datetime.datetime`]
        The date when the GIF is published.
    has_audio: :class:`bool`
        Wheather the GIF has sound.
    width: :class:`int`
        The GIF's width.
    height: :class:`int`
        The GIF's height.
    likes: :class:`int`
        The amount of likes for the GIF.
    tags: List[:class:`str`]
        A list of tags for the GIF.
    verified: :class:`bool`
        Wheather the publisher of the GIF is a verified creator.
    views: Optional[:class:`int`]
        The amount of views for the GIF.
    duration: :class:`float`
        The GIF's duration in seconds.
    published: :class:`bool`
        Wheather the GIF is published.
    urls: :class:`URL`
        The different types of URLs for the GIF.
    username: :class:`str`
        The username of the publisher.
    type: :class:`int`

    avg_color: :class:`str`

    """

    __slots__ = (
        'id', 'create_date', 'has_audio', 'width', 'height',
        'likes', 'tags', 'verified', 'views', 'duration',
        'published', 'urls', 'username', 'type', 'avg_color',
    )

    id: str
    create_date: Optional[datetime.datetime]
    has_audio: bool
    width: int
    height: int
    likes: int
    tags: List[str]
    verified: bool
    views: Optional[int]
    duration: float
    published: bool
    urls: URL
    username: str
    type: int
    avg_color: str
    # gallery

@dataclass
class Image:
    # TODO: Document "type" and "avg_color"
    """The image returned from RedGifs.

    Attributes
    ----------
    id: :class:`str`
        The image ID.
    create_date: Optional[:class:`datetime.datetime`]
        The date when the image is published.
    width: :class:`int`
        The image width.
    height: :class:`int`
        The image height.
    likes: :class:`int`
        The amount of likes the image has.
    tags: List[:class:`str`]
        A list of tags for the GIF.
    verified: :class:`bool`
        Wheather the publisher of the image is a verified creator.
    views: Optional[:class:`int`]
        The amount of views the image has.
    published: :class:`bool`
        Wheather the image is published.
    urls: :class:`URL`
        The different types of URLs for the image.
    username: :class:`str`
        The username of the publisher.
    type: :class:`int`

    avg_color: :class:`str`

    """

    __slots__ = (
        'id', 'create_date', 'width', 'height',
        'likes', 'tags', 'verified', 'views',
        'published', 'urls', 'username', 'type', 'avg_color',
    )

    id: str
    create_date: Optional[datetime.datetime]
    width: int
    height: int
    likes: int
    tags: List[str]
    verified: bool
    views: Optional[int]
    published: bool
    urls: URL
    username: str
    type: int
    avg_color: str

@dataclass
class User:
    # TODO: Document "subscription"
    """The user's information returned from RedGifs.

    Attributes
    ----------
    creation_time: Optional[:class:`datetime.datetime`]
        The user's account creation time.
    description: Optional[:class:`str`]
        The user's description on their profile.
    followers: :class:`int`
        The user's amount of followers.
    following: :class:`int`
        The user's amount of following users.
    gifs: :class:`int`
        The user's total amount of GIFs published.
    name: Optional[:class:`str`]
        The user's name.
    profile_image_url: Optional[:class:`str`]
        The user's profile image URL.
    profile_url: :class:`str`
        The user's "profile URL". It is a URL that is seen on the profile set by the user.
        This is NOT the user's URL on ``redgifs.com``, see :attr:`.User.url` for that.
    published_collections: Optional[:class:`int`]
        The user's amount of published collections.
    published_gifs: :class:`int`
        The user's amount of published GIFs.
    status: Optional[:class:`str`]
        The user's status.
    subscription: :class:`int`

    url: :class:`str`
        The user's URL on ``redgifs.com``.
    username: :class:`str`
        The user's username.
    verified: :class:`bool`
        Wheather the user is a verified creator.
    views: :class:`int`
        The user's total amount of views of all the GIFs published.
    poster: Optional[:class:`str`]
        The user's poster URL.
    preview: Optional[:class:`str`]
        The user's preview URL.
    thumbnail: Optional[:class:`str`]
        The user's thumbnail URL.
    links: Optional[List[Dict[:class:`str`, :class:`str`]]]
        The linked websites on the user's profile.
    """

    __slots__ = (
        'creation_time', 'description', 'followers', 'following', 'gifs',
        'name', 'profile_image_url', 'profile_url', 'published_collections', 'published_gifs',
        'status', 'subscription', 'url', 'username', 'verified',
        'views', 'poster', 'preview', 'thumbnail', 'links'
    )

    creation_time: Optional[datetime.datetime]
    description: Optional[str]
    followers: int
    following: int
    gifs: int
    name: str
    profile_image_url: Optional[str]
    profile_url: str
    published_collections: Optional[int]
    published_gifs: int
    status: Optional[str]
    subscription: int
    url: str
    username: str
    verified: bool
    views: int
    poster: Optional[str]
    preview: Optional[str]
    thumbnail: Optional[str]
    links: Optional[List[Dict[str, str]]]

@dataclass
class SearchResult:
    # TODO: Document "users"
    """The result you have searched. This is returned in :meth:`~redgifs.API.search()`.

    Attributes
    ----------
    searched_for: :class:`str`
        The result of what you have searched for.
        This may differ from what you have provided for ``query`` in :meth:`~redgifs.API.search()`.
    page: :class:`int`
        The current page number.
    pages: :class:`int`
        The total number of pages for the query.
    total: :class:`int`
        The total number of GIFs for the query.
    gifs: Optional[List[:class:`GIF`]]
        The GIFs which was searched for.
    images: Optional[List[:class:`Image`]]
        The images which was searched for.
    users: List[:class:`User`]

    tags: List[:class:`str`]
        The tags related to the GIFs and search query.
    """

    __slots__ = (
        'searched_for', 'page', 'pages', 'total', 'gifs', 'images', 'users', 'tags',
    )

    searched_for: str
    page: int
    pages: int
    total: int
    gifs: Optional[List[GIF]]
    images: Optional[List[Image]]
    users: List[User]
    tags: List[str]

@dataclass
class CreatorsResult:
    # TODO: Document "total"
    """The creator results searched for.

    Attributes
    ----------
    items: List[:class:`User`]
        The list of creators.
    page: :class:`int`
        The current page number.
    pages: :class:`int`
        The total number of pages available.
    total: :class:`int`

    """

    __slots__ = ('items', 'page', 'pages', 'total')

    items: List[User]
    page: int
    pages: int
    total: int

@dataclass
class CreatorResult:
    """The creator result searched for.

    Attributes
    ----------
    creator: :class:`User`
        The creator/user details.
    page: :class:`int`
        The current page number.
    pages: :class:`int`
        The total number of pages available.
    total: :class:`int`
        The total number of GIFs this creator/user has created.
    gifs: List[:class:`GIF`]
        The GIFs uploaded by this creator.
    images: List[:class:`Image`]
        The images uploaded by this creator.
    """

    __slots__ = ('creator', 'page', 'pages', 'total', 'gifs', 'images')

    creator: User
    page: int
    pages: int
    total: int
    gifs: List[GIF]
    images: List[Image]

class TagSuggestion(TypedDict):
    """The tag suggestion results.
    
    Attributes
    ----------
    name: :class:`str`
        The tag name.
    count: :class:`int`
        The number of GIFs with this tag.
    """
    name: str
    count: int



"""
The MIT License (MIT)

Copyright (c) 2022-present scrazzz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import re
import sys
import platform
import itertools
from pathlib import Path
from typing import Iterable, Optional, TYPE_CHECKING

import click
import yarl

import redgifs
from redgifs.enums import MediaType
from . import __version__

if TYPE_CHECKING:
    from redgifs.models import GIF, Image

def download_gif(client, url: yarl.URL, quality: str, folder: Optional[Path], *, skip_check: bool = False):
    # If skip_check is true then this will be the GIF's ID and splitting the URL is not required
    id = str(url).lower() if skip_check else url.path.split('/')[-1]
    gif = client.get_gif(id)
    gif_url = gif.urls.sd if quality == 'sd' else gif.urls.hd
    filename = f'{gif_url.split("/")[3].split(".")[0]}.mp4'
    dir_ = f'{folder}/{filename}' if folder else filename
    client.download(gif_url, dir_)
    click.echo(f'Downloading {id}...')
    click.echo('Download complete.')

def _dl_with_args(client, gif: GIF | Image, quality: str, folder: Optional[Path], is_image: bool):
    gif_url = gif.urls.sd if quality == 'sd' else gif.urls.hd
    filename = f'{gif_url.split("/")[3].split(".")[0]}.mp4'
    if is_image:
        name, ext = gif_url.split("/")[3].split('.')
        filename = f'{name}.{ext}'

    if folder:
        client.download(gif_url, f'{folder}/{filename}')
    else:
        client.download(gif_url, f'{filename}')

def download_users_gifs(client, url: yarl.URL, quality: str, folder: Optional[Path], images_only: bool):
    match = re.match(r'https://(www\.)?redgifs\.com\/users\/(?P<username>\w+)', str(url))
    if not match:
        click.UsageError(f'Not a valid redgifs user URL: {url}')
        exit(1)

    is_image = images_only

    user = match.groupdict()['username']
    data = client.search_creator(user, media_type=MediaType.IMAGE if is_image else MediaType.GIF)
    curr_page = data.page
    total_pages = data.pages
    total_gifs_in_page = data.gifs if not is_image else data.images
    total = data.total
    done = 0

     # Case where there is only 1 page
    if curr_page == total_pages:
        spinner = itertools.cycle(["-", "\\", "|", "/"])
        for gif in total_gifs_in_page:
            try:
                _dl_with_args(client, gif, quality, folder, is_image)
                done += 1
                click.echo(f'\r{next(spinner)} Downloaded {done}/{total} {"GIFs" if not is_image else "image"}', nl=False)
            except Exception as e:
                click.UsageError(f'[!] An error occurred when downloading {url}:\n{e}. Continuing...')
                continue

        click.echo(f'\r[-] Downloaded {done}/{total} {"videos" if not is_image else "images"} of user {user} {f"to folder '{folder}'" if folder else ""} sucessfully!')
        exit(0)

    # Case where there is more than 1 page
    while curr_page <= total_pages:
        spinner = itertools.cycle(["-", "\\", "|", "/"])
        for gif in total_gifs_in_page:
            try:
                _dl_with_args(client, gif, quality, folder, is_image)
                done += 1
                click.echo(f'\r{next(spinner)} Downloaded {done}/{total} {"GIFs" if not is_image else "image"}', nl=False)
            except Exception as e:
                click.echo(f'[!] An error occurred when downloading {url}:\n{e}. Continuing...')
                continue

        # If we are in the last page, break the loop
        if curr_page == total_pages:
            break

        curr_page += 1
        total_gifs_in_page.clear()
        data = client.search_creator(user, page=curr_page)
        total_gifs_in_page.extend(data.gifs)

    click.echo(f'\r[-] Downloaded {done}/{total} {"videos" if not is_image else "images"} of user {user} {f"to folder '{folder}'" if folder else ""} sucessfully!')

@click.command()
@click.argument('urls', nargs=-1)
@click.option('-v', '--version', is_flag=True, help='Shows currently installed version.')
@click.option('-q', '--quality', type=click.Choice(['sd', 'hd']), default='hd', show_default=True, help='Video quality of GIF to download.')
@click.option(
    '-f', '--folder',
    help='The folder to save the downloads to.',
    type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=Path),
    metavar='FOLDER_NAME'
)
@click.option(
    '-i', '--input', 'file',
    help='Download URLs from a newline seperated txt file.',
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    metavar='FILE_NAME'
)
@click.option('--images', is_flag=True, help='Download only images from a user profile.')
@click.pass_context
def cli(ctx: click.Context, urls: Iterable[str], quality: str, folder: Optional[Path], file: Optional[str], version: bool, images: bool) -> None:
    if version:
        info = []
        info.append('- python: v{0.major}.{0.minor}.{0.micro}-{0.releaselevel}'.format(sys.version_info))
        info.append('- redgifs: v{0.major}.{0.minor}.{0.micro}-{0.releaselevel}'.format(redgifs.version_info))
        info.append('- system info: {0.system} {0.release} {0.version}'.format(platform.uname()))
        return click.echo('\n'.join(info))

    if not urls and not file:
        return click.echo(ctx.command.get_help(ctx))

    client = redgifs.API().login()

    if file:
        with open(file) as f:
            [download_gif(client, url, quality, folder) for url in [yarl.URL(line) for line in f.readlines()]]

    for url in urls:
        url = yarl.URL(url)
        if 'redgifs' not in str(url.host):
            download_gif(client, url, quality, folder, skip_check=True)

        # Handle 'normal' URLs, i.e, a direct link (eg: "https://redgifs.com/watch/deeznuts")
        if '/watch/' in url.path:
            download_gif(client, url, quality, folder)

        # Handle /users/ URLs (eg: https://redgifs.com/users/redgifs)
        if '/users/' in url.path:
            download_users_gifs(client, url, quality, folder, images)
